/**
 *	Written 2012 by Tyler Montbriand.  Freeware.
 *	http://burningsmell.org/sensor8/
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "i2c.h"
#include "tcgets.h"
#include "tsm_getopt.h"
#include "sensor20.h"
#include "mydebug.h"

pin_t lookup_pin(const char *str, const char *tab[])
{
	int pin=0;
	while(tab[pin])
	{
		if(strcmp(str, tab[pin])==0) return(pin);
		pin++;
	}

	return(-1);
}

struct
{
	const char *dev;	// What dev to open, i.e. /dev/ttyUSB0
	const char *mode;	// What serial mode, probably 115200N81
	int verbose;		// Prints verbose debugging output when enabled.
	int usage;		// Prints usage when enabled.
	int raw;	// Prints raw bytes when enabled.
	int lock;	// Uses POSIX advisor locking when enabled.
	int i2cbytes;

	pin_t SCL, SDA;

	float vref;	// values 0-1023 are returned by ADC.  This scales to 0-5V by default
			// but an external voltage can be referenced so can be specified by
			// --vref instead.
} globals={
	.dev	=	"/dev/ttyUSB0",
	.mode	=	"115200N81",
	.verbose=	1, /* FIXME, should be 0 	*/
	.usage	=	0,
	.raw	=	0,
	.lock	=	0,
	.vref	=	5.0,
	.i2cbytes=	1,

	.SCL	=	PIN_FAIL,
	.SDA	=	PIN_FAIL,
	};

const struct opt mopts[];

int open_dev(void), close_dev(int fd);

/**
 *	Program to send commands to a DLP-IO20 USB I/O module.
 */
int main(int argc, char *argv[])
{
	int fd=-1, arg;

	if(getenv("SENSOR20"))
		globals.dev=getenv("SENSOR20");

	arg=tsm_getopt(argc, argv, mopts);
	if((arg < 0) || globals.usage)
	{
		fprintf(stderr, 
			"SENSOR20, utility to talk to a DLP-IO20 module\n"
			"Written 2012 Tyler Montbriand.  Freeware.\n\n");
		tsm_usage(argc, argv, mopts);
		fprintf(stderr, "\nValid commands (PIN can be AN0-AN13, RA4, P5-P7, RB6-RB7)\n"
			"\tPIN\t\tLogic value on any pin, ex AN7, reads logic value on AN7\n"
			"\tPIN:T\t\tRead DS18b20 from any pin, ex AN7:T, reads temp from AN7\n"
			"\tPIN:F\t\tLike :T, but in Fahrenheit\n"
			"\tPIN:'?'\t\tDetect serial number of DS18b20 from any pin, ex AN7:'?'\n"
			"\tPIN:[0-1]\tSet logic high or low on any pin, ex AN7:1\n"
			"\tPIN:Z\t\tSet tri-state on any pin, ex AN7:Z\n"
			"\tAN[0-13]:V\tex AN7, reads ADC voltage from AN7\n"
			"\tAN[0-13]:M(1K|2K|4K|8K|20K):(128|256|512|1024|2048|4096|8192)\n"
			"\t\t\tCapture a bunch of voltage samples at a given rate.\n"
			"\t\t\ti.e. AN7:4K:512, capture 512 samples at 4K samples/s.\n"
			"\tRB[6-7]:R\tEnable and reset pulse counts on RB6/RB7\n"
			"\tRB[6-7]:C\tPrint pulse counts from RB6/RB7\n"
			"\tFLASH\t\tCauses the built-in LED to flash briefly, then stay on\n"
			"\tLED:[0-1]\tTurns the built-in LED on and off\n"
			"\tR:[1-2]:[0-1]\tex R:2:1, Reset contacts of relay 2\n"
			"\tD:[0-9]*\tex D:100, delay 100 milliseconds\n"
			"\tU:[0-9]*\tex U:1000, delay 1000 microseconds\n"
			"\tRAWI2C:datapin:clockpin:command,command2,command3,...\n"
			"\t\tRaw I2C sequence.  You can put start/stop/nack wherever\n"
			"\t\tyou feel appropriate.  Wordy but should be capable of anything.\n"
			"\t\t\tR\t\tRead command\n"
			"\t\t\tW\t\tWrite command\n"
			"\t\t\tA\t\t7-bit address\n"
			"\t\t\tS\t\tSend start bit\n"
			"\t\t\tP\t\tSend stop bit\n"
			"\t\t\tN\t\tSend NACK\n"
			"\t\t\txx\t\tTwo hexadecimal digits\n"
			"\t\tExample:  Read device ID(register 00) from OSEPP accelerometer(addr 1D)\n"
			"\t\tRAWI2C:AN5:AN6:SAW1D,00,SAR1D,NRP\n"
			"\tI7:datapin:clockpin:deviceid[register]\n"
			"\t\t7-bit address read of deviceid and reg, i.e. I7R:AN5:AN6:1d[00]\n"
			"\tI7:datapin:clockpin:deviceid[register]=xx\n"
			"\t\t7-bit address write of device and register, i.e. I7W:AN5:AN6:1d[1e]=01\n",
			argv[0] );
		return(1);
	}

	if(arg == argc)
	{
		fprintf(stderr, "No commands given\n");
		return(2);
	}

	if(globals.verbose)	setenv("DEBUG", "1", 1);

	fd=open_dev();

	if(fd < 0)
	{
		perror(globals.dev);
		return(1);
	}

#ifdef USE_LOCK
	if(globals.lock)
	{
		struct flock l;

		memset(&l, 0, sizeof(l));
		l.l_type=F_WRLCK;
		l.l_whence=SEEK_SET;
		l.l_start=0L;
		l.l_len=512L;

		if(fcntl(fd, F_SETLKW, &l) != 0)
		{
			perror("Couldn't lock");
			close(fd);
			return(1);
		}
	}
#endif

	arg--;
	while((++arg) < argc)
	{
		int devid,reg, byte=-1;
		char spin1[64], spin2[64];
		char buf[256], *arglist[64], buf2[64];
		int n, m, ret;
		pin_t pin;

		strcpy(buf, argv[arg]);

		arglist[0]=strtok(buf, ":");
		for(n=1; n<64; n++)
		{
			arglist[n]=strtok(NULL, ":");
			if(arglist[n] == NULL) arglist[n]="";
		}

		cmd20_err=0;

		if(strcmp(arglist[0], "PING") == 0)
		{
			if(cmd20_ping(fd) < 0)
			if(cmd20_ping(fd) < 0)
			{
				fprintf(stderr, "Could not contact device\n");
				exit(1);
			}

			fprintf(stderr, "Device contacted\n");
		}
		else if(strcmp(argv[arg], "FLASH") == 0)	ret=cmd20_flash(fd);
		else if(strcmp(argv[arg], "INTREF") == 0)	ret=cmd20_internalref(fd);
		else if(strcmp(argv[arg], "EXTREF") == 0)	ret=cmd20_externalref(fd);
		else if(sscanf(argv[arg],  "LED:%d", &n) == 1)	ret=cmd20_ledctl(fd, n);
		else if(sscanf(argv[arg],  "R:%d:%d",&n,&m)==2)	ret=cmd20_relay(fd, n, m);
		else if(sscanf(argv[arg],  "D:%d", &n) == 1)	usleep(1000*n);
		else if(sscanf(argv[arg],  "U:%d", &n) == 1)	usleep(n);
		else if(strcmp(arglist[0],  "SCL")==0)
			globals.SCL=lookup_pin(arglist[1],pintab);
		else if(strcmp(arglist[0],  "SDA")==0)
			globals.SDA=lookup_pin(arglist[1],pintab);
		else if(strcmp(arglist[0], "RAWI2C") == 0)
		{
			pin_t pin2=lookup_pin(arglist[2], pintab);
			pin=lookup_pin(arglist[1], pintab);

			if((pin2 == PIN_FAIL) || (pin == PIN_FAIL))
			{
				ret=-1;
				cmd20_err=CMD20_ERR_PARAM;
				fprintf(stderr, "couldn't find pin(%s,%s)\n", arglist[1], arglist[2]);
				goto CMD_FAIL;
			}

			ret=cmd20_i2c(fd, pin, pin2, arglist[3]);
		}
		else if(sscanf(argv[arg], "B:%d", &globals.i2cbytes) == 1)
		{
			fprintf(stderr, "Set bytes to %d\n", globals.i2cbytes);
		}
		else if(sscanf(argv[arg], "I7:%[^:]:%[^:]:%02x[%02x]=%02x", spin1, spin2, &devid, &reg, &byte) > 3)
		{
			pin_t pin2=lookup_pin(spin2, pintab);
			pin=lookup_pin(spin1, pintab);

			if((pin2 == PIN_FAIL) || (pin == PIN_FAIL))
			{
				ret=-1;
				cmd20_err=CMD20_ERR_PARAM;
				goto CMD_FAIL;
			}

			if(byte >= 0)
				ret=cmd20_i2c_write7(fd, pin, pin2, devid, reg, byte);
			else	ret=cmd20_i2c_read7(fd, pin, pin2, devid, reg);
		}
		else if( (pin=lookup_pin(arglist[0], pintab)) != PIN_FAIL)
		{
			switch(arglist[1][0])
			{
			short int t;
			short unsigned int u;
			unsigned int lu;
			unsigned char serial[8];
			unsigned short mbuf[8192];

			case '\0':
				ret=cmd20_in(fd, pin, &n);
				if(ret == 0) printf("%d\n", n);
				break;
			case '?':
				ret=cmd20_18b20_detect(fd, pin, serial);
				if(ret >= 0)
				{
					for(n=0; n<8; n++) printf("%02x", serial[n]);
					printf("\n");
				}
				break;
			case 'Z':
				ret=CMD20_TRISTATE(fd, pin);
				break;
			case 'S':
				ret=cmd20_18b20_start(fd, pin);
				break;
			case 'T':
				ret=cmd20_18b20_read(fd, pin, &t);
				if(ret >= 0)	printf("%2.2f\n", t/16.0f);
				break;
			case 'F':
				ret=cmd20_18b20_read(fd, pin, &t);
				if(ret >= 0)	printf("%2.2f\n", (t/16.0)*(9.0/5.0)+32);
				break;

			case 'V':
				ret=cmd20_voltage(fd, pin, &u);
				if(ret >= 0)
					printf("%2.2f\n", (u*globals.vref)/1024.0);
				break;

			case '0':	case '1':
				ret=cmd20_out(fd, pin, arglist[1][0]-'0');
				break;
			case 'M':
				n=lookup_pin(arglist[1]+1, sample_rate);

				m=-1;
				sscanf(arglist[2], "%d", &m);

				for(u=0; sample_size[u] > 0; u++)
					if(sample_size[u] == m) break;

				if((sample_size[u] < 0) || (n<0))
				{
					ret=-1;
					cmd20_err=CMD20_ERR_PARAM;
					break;
				}

				ret=cmd20_multivoltage(fd, pin, n, u, mbuf);
				if(ret >= 0)
				for(n=0; n<sample_size[u]; n++)
					printf("%2.2f\n", (mbuf[n]*globals.vref)/1024.0);

				break;
			case 'R':				
				if(sscanf(arglist[2], "%d", &n) != 1)
				{
					ret=-1;
					cmd20_err=CMD20_ERR_PARAM;
					break;
				}

				ret=cmd20_enablectr(fd, pin, n);
				break;
			case 'C':
				ret=cmd20_readctr(fd, pin, &lu);
				if(ret >= 0) printf("%d\n", lu);
				break;
			default:
				goto CMD_UNKNOWN;
			}
		}
		else
		{
CMD_UNKNOWN:
			fprintf(stderr, "Unknown command '%s'\n", argv[arg]);
			ret=-1;
		}

CMD_FAIL:
		if(ret < 0)
			fprintf(stderr, "Command failed:  '%s'\n", argv[arg]);
		if(cmd20_err > 0)
			fprintf(stderr, "Error was:  %s\n", cmd20_errors[cmd20_err]);
	}

#ifdef USE_LOCK
	if(globals.lock)
	{
		struct flock l;

		memset(&l, 0, sizeof(l));
		l.l_type=F_UNLCK;
		l.l_whence=SEEK_SET;
		l.l_start=0L;
		l.l_len=512L;

		if(fcntl(fd, F_SETLK, &l) != 0)
		{
			perror("Couldn't clear lock");
			close_dev(fd);
			return(1);
		}
	}
#endif

	close_dev(fd);
	return(0);
}

int close_dev(int fd)
{
	if(fd >= 0)
	{
		if(strcmp(globals.mode, "-") != 0)	// Change no TTY settings
		if(termios_restore(fd) < 0)
		fprintf(stderr, "Couldn't restore termios\n");

		close(fd);
	}
}

int open_dev(void)
{
	int fd=-1;

	fd=open(globals.dev, O_RDWR);

	if(fd < 0)
		return(-1);

	if(strcmp(globals.mode, "-") == 0)	// Change no TTY settings
		return(fd);

	if(termios_save(fd) < 0)
	{
		close(fd);
		return(-1);
	}

	if(termios_set(fd, globals.mode) < 0)
	{
		close(fd);
		return(-1);
	}

	return(fd);
}

const struct opt mopts[]={
	{"device",	'd',	1, &(globals.dev), NULL, "-d /dev/ttyUSB0" },
	{"verbose",	'v',	0, &(globals.verbose), NULL, "Enables debug output" },
	{"help",	'h',	0, &(globals.usage), NULL, "Shows this listing" },
	{"raw",		'r',	0, &(globals.raw), NULL, "Outputs raw bytes"},
	{"lock",	'l',	0, &(globals.lock), NULL, "Enables cooperative locking" },
	{"vref",	0,	1, &(globals.vref), "%f", "\tVoltage reference for the ADC"},
	{NULL,		0,	0,	NULL,	 NULL, 	NULL }
	};
