/**
 *	Written 2013 by Tyler Montbriand.  Freeware.
 *	http://burningsmell.org/sensor8/
 */

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "i2c.h"
#include "tcgets.h"
#include "tsm_getopt.h"
#include "sensor20.h"
#include "mydebug.h"
#include "help.h"

pin_t lookup_pin(const char *str, const char *tab[]);

#ifndef NAN
#define NAN (0.0/0.0)
#endif

struct
{
	const char *dev;	// What dev to open, i.e. /dev/ttyUSB0
	const char *mode;	// What serial mode, probably 115200N81
	int verbose;		// Prints verbose debugging output when enabled.
	int usage;		// Prints usage when enabled.
	int lock;	// Uses POSIX advisory locking when enabled.
	int bug;	// FTDI kernel bug workaround
	// Int, not the enum types, since we need the signs
	int avg;	// Whether to oversample analog readings
	int SCL, SDA;
	int rate;

	float vref;	// values 0-1023 are returned by ADC.  When this is NAN, it assumes
			// a 5.0v internal reference.  If you set it to a value,
			// it will assume an external reference on AN3 is being used.
} globals={
	.dev	=	"/dev/ttyUSB0",
	.mode	=	"115200N81",
	.verbose=	0,
	.usage	=	0,
	.lock	=	1,
	.vref	=	NAN,
	.SCL	=	PIN_FAIL,
	.SDA	=	PIN_FAIL,
	.rate	=	RATE_20K,
	.bug=	0,
	.avg=	-1,
	};

const struct opt mopts[];

int open_dev(void), close_dev(int fd);

int compare_s(const short *a, const short *b)
{	return((*a)-(*b)); }

typedef int (*compare_func)(const void *a, const void *b);

/**
 *	Program to send commands to a DLP-IO20 USB I/O module.
 */
int main(int argc, char *argv[])
{
	int fd=-1, arg;

	if(getenv("SENSOR20"))	globals.dev=getenv("SENSOR20");
	if(getenv("VREF"))	sscanf(getenv("VREF"), "%f", &globals.vref);

	if(getenv("SCL"))	globals.SCL=lookup_pin(getenv("SCL"),pintab);
	if(getenv("SDA"))	globals.SDA=lookup_pin(getenv("SDA"),pintab);
	// TODO:  Same for rate, avg

	arg=tsm_getopt(argc, argv, 1, mopts);
	if((arg < 0) || globals.usage)
	{
		fprintf(stderr, "SENSOR20 v0.0.6, utility to talk to a DLP-IO20 module\n"
				"Written 2013 Tyler Montbriand.  Freeware.\n\n");

		tsm_usage(argc, argv, mopts);
		if(globals.verbose)
		{
			fputs(helptext, stderr);
			fprintf(stderr, "\nBroke your buffer?  Try %s 2>&1 | less\n", argv[0]);
		}

		return(1);
	}

	if(globals.verbose)	setenv("DEBUG", "1", 1);

	if(arg == argc)
	{
		fprintf(stderr, "No commands given\n");
		return(2);
	}

	if(globals.bug)
	{
		const char *old=globals.mode;
		globals.mode="115200N81";
		close(open_dev());
		globals.mode="115200N81H";
		close(open_dev());
		globals.mode=old;
	}

	fd=open_dev();

	if(fd < 0)
	{
		perror(globals.dev);
		return(1);
	}

#ifdef USE_LOCK
	if(globals.lock)
	{
		struct flock l;

		memset(&l, 0, sizeof(l));
		l.l_type=F_WRLCK;
		l.l_whence=SEEK_SET;
		l.l_start=0L;
		l.l_len=512L;

		if(fcntl(fd, F_SETLKW, &l) != 0)
		{
			perror("Couldn't lock");
			close(fd);
			return(1);
		}
	}
#endif

	if(!isnan(globals.vref))
	{
		INFO_PRINT("Enabling external reference on AN3\n");
		if(cmd20_externalref(fd) <0)
			perror("error using external reference");
	}

	while(arg < argc)
	{
		int devid,reg, byte=-1;
		char buf[256], *arglist[64], buf2[64], *last=NULL;
		int n, m, ret;
		pin_t pin;

		// Parse more commandline args
		if(argv[arg][0] == '-')
		{
			n=tsm_getopt(argc, argv, arg, mopts);
			if(n < 0)
				exit(1);

			arg=n;
			continue;
		}

		buf2[0]=0;

		// TODO:  optimize
		cmd20_i2c_restart();

		strcpy(buf, argv[arg]);

		arglist[0]=strtok(buf, ":[]=");
		for(n=1; n<64; n++)
		{
			arglist[n]=strtok(NULL, ":[]=");
			if(arglist[n]) last=arglist[n];
			else arglist[n]="";
		}

		cmd20_err=0;

		if(strcmp(arglist[0], "PING") == 0)
		{
			if(cmd20_ping(fd) < 0)
			if(cmd20_ping(fd) < 0)
			{
				INFO_PRINT("Could not contact device\n");
				exit(1);
			}

			INFO_PRINT("Device contacted\n");
		}
		else if(strcmp(argv[arg],  "FLASH") == 0)	ret=cmd20_flash(fd);
		else if(sscanf(argv[arg],  "EXTREF:%f", &globals.vref) == 1)
			ret=cmd20_externalref(fd);
		else if(strcmp(argv[arg],  "CALIBRATE") == 0)
		{
			unsigned short u;
			ret=cmd20_internalref(fd);
			cmd20_voltage(fd, AN3, &u);
			globals.vref=(u/1024.0)*5.0;
			INFO_PRINT("Referencing to %.2f\n", globals.vref);
		}
		else if(strcmp(argv[arg],  "INTREF") == 0)
		{
			ret=cmd20_internalref(fd);
			if(ret == 0)		globals.vref=NAN;
		}
		else if(sscanf(argv[arg],  "LED:%d", &n) == 1)	ret=cmd20_ledctl(fd, n);
		else if(sscanf(argv[arg],  "R:%d:%d",&n,&m)==2)	ret=cmd20_relay(fd, n, m);
		else if(sscanf(argv[arg],  "D:%d", &n) == 1)	ret=usleep(1000*n);
		else if(sscanf(argv[arg],  "U:%d", &n) == 1)	ret=usleep(n);
		else if(strcmp(arglist[0], "RAWI2C") == 0)
		{
			if((globals.SCL == PIN_FAIL) || (globals.SDA == PIN_FAIL))
			{
				ret=-1;
				cmd20_err=CMD20_ERR_PARAM;
				INFO_PRINT("couldn't find pin(%s,%s)\n", arglist[1], arglist[2]);
				goto CMD_FAIL;
			}

			ret=cmd20_i2c(fd, globals.SCL, globals.SDA, arglist[3]);
		}
		else if(sscanf(argv[arg], "%02x[%02x]=%02x", &devid, &reg, &byte) == 3)
		{
			unsigned char d[256], pos=0;
			if((globals.SCL == PIN_FAIL) || (globals.SDA == PIN_FAIL))
			{
				ret=-1;
				cmd20_err=CMD20_ERR_PARAM;
				goto CMD_FAIL;
			}

			while(sscanf(last, "%02hhx", d+(pos++)) == 1)	last += 2;

			pos--;

			ret=cmd20_i2c_write7_multi(fd, globals.SCL, globals.SDA, devid, reg, d, pos);
		}
		else if(sscanf(argv[arg], "%02x[]=%02x", &devid, &byte) == 2)	
		{
			unsigned char d[256], pos=0;
			if((globals.SCL == PIN_FAIL) || (globals.SDA == PIN_FAIL))
			{
				ret=-1;
				cmd20_err=CMD20_ERR_PARAM;
				goto CMD_FAIL;
			}

			while(sscanf(last, "%02hhx", d+(pos++)) == 1)	last += 2;

			pos--;

			ret=cmd20_i2c_write7raw(fd, globals.SCL, globals.SDA, devid, d, pos);
		}
		else if(sscanf(argv[arg], "%02x[%[]]", &devid, buf2) == 2)
		{
			if(sscanf(last, "%d", &byte) != 1) byte=1;
			ret=cmd20_i2c_read7raw(fd, globals.SCL, globals.SDA, devid, byte);
		}
		else if(sscanf(argv[arg], "%02x[%02x]:%02x", &devid, &reg, &byte) > 1)
		{
			if((globals.SCL == PIN_FAIL) || (globals.SDA == PIN_FAIL))
			{
				ret=-1;
				cmd20_err=CMD20_ERR_PARAM;
				goto CMD_FAIL;
			}

			if(byte < 1) byte=1;

			ret=cmd20_i2c_read7_multi(fd, globals.SCL, globals.SDA,	devid, reg, byte);
		}
		else if( (pin=lookup_pin(arglist[0], pintab)) != PIN_FAIL)
		{
			float ref=globals.vref;
			int x=-1;

			if(arglist[1][0] == '\0') // Blank
			{
				ret=cmd20_in(fd, pin, &n);
				if(ret == 0) printf("%d\n", n);
			}
			else while(arglist[1][++x])
			switch(toupper(arglist[1][x]))
			{
			short int t;
			short unsigned int u;
			unsigned int lu;
			unsigned char serial[8];
			unsigned short mbuf[8192];
			const char *cmd;
			case 'I':
				ret=cmd20_in(fd, pin, &n);
				if(ret == 0) printf("%d\n", n);
				break;

			case '?':
				ret=cmd20_18b20_detect(fd, pin, serial);
				if(ret >= 0)
				{
					for(n=0; n<8; n++) printf("%02x", serial[n]);
					printf("\n");
				}
				break;
			case 'Z':	ret=CMD20_TRISTATE(fd, pin);	break;
			case 'S':	ret=cmd20_18b20_start(fd, pin);	break;
			case 'T':	case 'F':
				ret=cmd20_18b20_read(fd, pin, &t);
				if(ret < 0)
					break;

				if(toupper(arglist[1][x]) == 'F')
					printf("%2.2f\n", (t/16.0)*(9.0/5.0)+32);
				else	printf("%2.2f\n", t/16.0f);
				break;

			case 'V':
				if(globals.avg>=0)
				{
					cmd="%2.4f\n";
					ret=cmd20_oversample(fd, pin, globals.rate, globals.avg, &lu);
					t=sample_size[globals.avg];
				}
				else
				{
					cmd="%2.2f\n";
					ret=cmd20_voltage(fd, pin, &u);
					lu=u;
					t=1;
				}

				if(ret >= 0)
				{
					if(isnan(ref)) ref=5.0;
					printf(cmd, (lu*ref)/(1024.0*t));
				}
				break;

			case '0':	case '1':
				ret=cmd20_out(fd, pin, arglist[1][x]-'0');
				break;
			case 'M':
				n=lookup_pin(arglist[1]+1, sample_rate);

				m=-1;
				sscanf(arglist[2], "%d", &m);

				for(u=0; sample_size[u] > 0; u++)
					if(sample_size[u] == m) break;

				if((sample_size[u] < 0) || (n<0))
				{
					ret=-1;
					cmd20_err=CMD20_ERR_PARAM;
					break;
				}

				ret=cmd20_multivoltage(fd, pin, n, u, mbuf);
				if(ret >= 0)
				for(n=0; n<sample_size[u]; n++)
					printf("%2.2f\n", (mbuf[n]*globals.vref)/1024.0);
				break;
			case 'R':
				if(arglist[1][x] == 'R')
					ret=cmd20_enablectr(fd, pin, 1);
				else	ret=cmd20_enablectr(fd, pin, 0);
				break;
			case 'C':
				ret=cmd20_readctr(fd, pin, &lu);
				if(ret >= 0) printf("%d\n", lu);
				break;
			default:
				goto CMD_UNKNOWN;
			}
		}
		else
		{
CMD_UNKNOWN:
			INFO_PRINT("Unknown command '%s'\n", argv[arg]);
			ret=-1;
		}

CMD_FAIL:
		if(ret < 0)
		{
			INFO_PRINT("Command failed:  '%s'\n", argv[arg]);
			printf("ERR\n");
		}
		if(cmd20_err > 0)
			fprintf(stderr, "Error was:  %s\n", cmd20_errors[cmd20_err]);

		arg++;
	}

#ifdef USE_LOCK
	if(globals.lock)
	{
		struct flock l;

		memset(&l, 0, sizeof(l));
		l.l_type=F_UNLCK;
		l.l_whence=SEEK_SET;
		l.l_start=0L;
		l.l_len=512L;

		if(fcntl(fd, F_SETLK, &l) != 0)
		{
			perror("Couldn't clear lock");
			close_dev(fd);
			return(1);
		}
	}
#endif

	close_dev(fd);
	return(0);
}

int close_dev(int fd)
{
	if(fd >= 0)
	{
		if(strcmp(globals.mode, "-") != 0)	// Change no TTY settings
		if(termios_restore(fd) < 0)
		fprintf(stderr, "Couldn't restore termios\n");

		close(fd);
	}
}

int open_dev(void)
{
	int fd=-1;

	fd=open(globals.dev, O_RDWR);

	if(fd < 0)
		return(-1);

	if(strcmp(globals.mode, "-") == 0)	// Change no TTY settings
		return(fd);

	if(termios_save(fd) < 0)
	{
		close(fd);
		return(-1);
	}

	if(termios_set(fd, globals.mode) < 0)
	{
		close(fd);
		return(-1);
	}

	return(fd);
}

const struct opt mopts[]={
	{"help",	'h',	0, &(globals.usage), NULL, "Shows this listing" },
	{"verbose",	'v',	0, &(globals.verbose), NULL, "Debug output and better help"},
	{"bug",		0,	0, &globals.bug, NULL, "Workaround FTDI kernel bug" },
	{"device",	'd',	1, &(globals.dev), NULL, "-d /dev/ttyUSB0" },
	{"rate",	'r',	1, &(globals.rate), NULL, "Sample rate 1K ... 20K", sample_rate },
//	{"raw",		'r',	0, &(globals.raw), NULL, "Outputs raw bytes"},
	{"avg",		'a',	1, &(globals.avg), "%d", "Multi-sample analog average", sample_size_c},
	{"lock",	'l',	0, &(globals.lock), NULL, "Enables cooperative locking" },
	{"vref",	0,	1, &(globals.vref), "%f", "\tVoltage reference for the ADC"},
	{"SDA",		0,	1, &(globals.SDA), NULL, "\tI2C data pin", pintab},
	{"SCL",		0,	1, &(globals.SCL), NULL, "\tI2C clock pin", pintab},
	{NULL,		0,	0,	NULL,	 NULL, 	NULL }
	};

pin_t lookup_pin(const char *str, const char *tab[])
{
	int pin=0;
	if(str == NULL) return(-1);

	while(tab[pin])
	{
		if(strcmp(str, tab[pin])==0) return(pin);
		pin++;
	}

	return(-1);
}
