#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <arpa/inet.h>
#include <stdint.h>

#include "sensor20.h"
#include "mydebug.h"

#define FAIL(X)		{	cmd20_err=X;	return(-1);		}
#define SUCCESS()	{	cmd20_err=CMD20_ERR_NONE; return(0);	}

static const unsigned char blank_serial[8]={0,0,0,0,0,0,0,0};

static unsigned char b20_serial[PIN_MAX][8]={
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0} };
static const union byteorder {
	unsigned char bytes[4];
	unsigned int word;
} little_endian={ { 0xef, 0xbe, 0xad, 0xde } };

int cmd20_err=0;

static const unsigned int native_order=0xdeadbeef;
#define IS_LITTLE_ENDIAN	(native_order == little_endian.word)

// Note the values are very wrong compared to the order all the other commands use.
// So we just use this internally, and convert from pin_t
enum {
	B2_CHANNEL_RB6	=	6,
	B2_CHANNEL_RB7	=	7,
};

static int cmd20_transact(int fd, unsigned char *reply, int rbytes, ...);

int cmd20_externalref(int fd)
{	return(cmd20_transact(fd, NULL, 0, CMD_EXTERNALREF, -1));	}

int cmd20_internalref(int fd)
{	return(cmd20_transact(fd, NULL, 0, CMD_INTERNALREF, -1));	}

int cmd20_oversample(int fd, pin_t pin, rate_t rate, samples_t samples, 
	unsigned int *total)
{
	int n, ret;
	unsigned short int v[8192];

	ret=cmd20_multivoltage(fd, pin, rate, samples, v);
	if(ret < 0) return(ret);

	*total=0;
	for(n=0; n<sample_size[samples]; n++) (*total) += v[n];
	SUCCESS();
}

int cmd20_multivoltage(int fd, pin_t pin, rate_t rate, samples_t samples, unsigned short *arr)
{
	int n;

	if((pin < AN0) || (pin>AN13))				FAIL(CMD20_ERR_PARAM);
	if((rate < RATE_1K) || (rate > RATE_20K))		FAIL(CMD20_ERR_PARAM);
	if((samples < SAMPLES_128) || (samples>SAMPLES_8192))	FAIL(CMD20_ERR_PARAM);

	DEBUG_PRINT(("rate %s samples %d\n", sample_rate[rate], sample_size[samples]));

	if(cmd20_transact(fd, (unsigned char *)arr, 
		sample_size[samples]*2, CMD_MULTIREAD, pin, rate, samples, -1) < 0)
		return(-1);

	if(!IS_LITTLE_ENDIAN)	for(n=0; n<sample_size[samples]; n++)	arr[n]=ntohs(arr[n]);

	SUCCESS();
}

int cmd20_voltage(int fd, pin_t pin, unsigned short *result)
{
	if((pin < AN0) || (pin>AN13)) FAIL(CMD20_ERR_PARAM);

	if(cmd20_transact(fd, (unsigned char *)result, 2, CMD_ANALOGREAD, pin, -1) < 0)
		return(-1);

	if(!IS_LITTLE_ENDIAN) (*result)=ntohs(*result);
	SUCCESS();
}

int cmd20_check_serial(int fd, int pin)
{
	if(memcmp(b20_serial[pin], blank_serial, 8) == 0)
	{
		if(cmd20_transact(fd, b20_serial[pin], 8,
			CMD_DETECTSENSOR, pin, -1) < 0)
			return(-1);
		if(memcmp(b20_serial[pin], blank_serial, 8)==0)
			FAIL(CMD20_ERR_NOSENSOR);


		if(b20_serial[pin][0] == 1)	FAIL(CMD20_ERR_SHORT);
		if(b20_serial[pin][0] == 2)	FAIL(CMD20_ERR_NOSENSOR);

		INFO_PRINT("Detected %02x%02x%02x%02x%02x%02x%02x%02x on %s\n",
			b20_serial[pin][0],	b20_serial[pin][1],
			b20_serial[pin][2],	b20_serial[pin][3],
			b20_serial[pin][4],	b20_serial[pin][5],
			b20_serial[pin][6],	b20_serial[pin][7],
			pintab[pin]);
	}

	if(b20_serial[pin][0] == 1)	FAIL(CMD20_ERR_SHORT);
	if(b20_serial[pin][0] == 2)	FAIL(CMD20_ERR_NOSENSOR);

	return(0);
}

int cmd20_18b20_setresolution(int fd, pin_t pin, int bits)
{
	if((bits < 9) || (bits > 12))		FAIL(CMD20_ERR_PARAM);
	if((pin < 0) || (pin > RB6))		FAIL(CMD20_ERR_PARAM);
	if(cmd20_check_serial(fd, pin) < 0)	FAIL(CMD20_ERR_NODETECT);
	return(cmd20_transact(fd, NULL, 0, CMD_SETRES, pin, bits, -1));
}

int cmd20_18b20_read(int fd, pin_t pin, short *temp)
{
	union {
		short word;
		unsigned short uword;
		unsigned char bytes[2];
	} d;

	if((pin < 0) || (pin > RB6))		FAIL(CMD20_ERR_PARAM);
	if(cmd20_check_serial(fd, pin) < 0)	FAIL(CMD20_ERR_NODETECT);

	if(cmd20_transact(fd, d.bytes, 2, CMD_READSENSOR, pin, -1)<0)
		return(-1);

	if(!IS_LITTLE_ENDIAN) d.uword=ntohs(d.uword);
	if(temp) (*temp)=d.word;

	// TODO:  check for suspicious TEMP_MIN or TEMP_MAX
	if(	(d.uword == TEMP_ERROR)	||
		(d.uword == TEMP_RESET)	||
		(d.word > TEMP_MAX)	||
		(d.word < TEMP_MIN)	||
		(d.uword == 0)	)
		FAIL(CMD20_ERR_NOTREADY);

	SUCCESS();
}

int cmd20_18b20_start(int fd, pin_t pin)
{
	if((pin < 0) || (pin > RB6))	FAIL(CMD20_ERR_PARAM);
	if(cmd20_check_serial(fd, pin) < 0)	FAIL(CMD20_ERR_NODETECT);
	return(cmd20_transact(fd, NULL, 0, CMD_CONVERTSENSOR, pin, -1));
}

int cmd20_18b20_detect(int fd, pin_t pin, unsigned char serial[8])
{
	if((pin < 0) || (pin > RB6))	return(-1);

	if(cmd20_transact(fd, b20_serial[pin], 8, CMD_DETECTSENSOR, pin, -1) < 0)
		return(-1);

	if(cmd20_check_serial(fd,pin) < 0)	return(-1);

	memcpy(serial, b20_serial[pin], 8);

	SUCCESS();
}

int cmd20_readctr(int fd, pin_t pin, unsigned int *value)
{
	if(pin == RB6)	pin=B2_CHANNEL_RB6;
	else if(pin == RB7)	pin=B2_CHANNEL_RB7;
	else	FAIL(CMD20_ERR_PARAM);

	if(cmd20_transact(fd, (unsigned char *)value, 4, CMD_READCTR, pin, -1) < 0)
		return(-1);

	if(!IS_LITTLE_ENDIAN) (*value)=ntohl(*value);

	SUCCESS();
}

int cmd20_enablectr(int fd, pin_t pin, int edge)
{
	if(pin == RB6)	pin=B2_CHANNEL_RB6;
	else if(pin == RB7)	pin=B2_CHANNEL_RB7;
	else	FAIL(CMD20_ERR_PARAM);

	return(cmd20_transact(fd, NULL, 0, CMD_ENABLECTR, pin, edge&1, -1));
}

int cmd20_opendrain(int fd, pin_t pin, int value, int *result)
{
	unsigned char buf;

	if(!(	((pin >= AN0) && (pin <= AN13)) ||
		(pin == RB6) || (pin == RB7) ))
		FAIL(CMD20_ERR_PARAM);

	(*result)=0;
	if(value)
	{
		int err=cmd20_transact(fd, &buf, 1, CMD_IOCMD, pin, B3_INPUT, 0, -1);
		(*result)=buf;
		return(err);
	}

	return(cmd20_transact(fd, NULL, 0, CMD_IOCMD, pin, B3_OUTPUT, 0, -1));
}

int cmd20_in(int fd, pin_t pin, int *value)
{
	unsigned char buf;
	if((pin < 0) || (pin > RB6))	FAIL(CMD20_ERR_PARAM);

	if(cmd20_transact(fd, &buf, 1, CMD_IOCMD, pin, B3_INPUT, 0, -1) < 0)
		return(-1);

	if(value != NULL) (*value)=buf;
	SUCCESS();
}

int cmd20_out(int fd, pin_t pin, int value)
{
	if((pin < 0) || (pin > RB6))	FAIL(CMD20_ERR_PARAM);

	return(cmd20_transact(fd, NULL, 0, CMD_IOCMD, pin, B3_OUTPUT, value&1, -1));
}

int cmd20_flash(int fd)
{	return(cmd20_transact(fd, NULL, 0, CMD_FLASHLED, -1));		}

int cmd20_ledctl(int fd, int led)
{	return(cmd20_transact(fd, NULL, 0, CMD_LEDCTL, (led&1)^1, -1));	}

int cmd20_relay(int fd, int which, int where)
{
	if((which < B2_RELAY_1) || (which > B2_RELAY_2))
		FAIL(CMD20_ERR_PARAM);

	return(cmd20_transact(fd, NULL, 0, CMD_RELAYCTL, which, where&1, -1));
}

int cmd20_ping(int fd)
{
	unsigned char reply;
	if(cmd20_transact(fd, &reply, 1, CMD_PING, -1) != 0)
		return(-1);
	if(reply != 'Y')
		return(-1);

	SUCCESS();
}

int cmd20_compile(unsigned char *buf, int *pos, int max, ...)
{
	va_list ap;
	int n;

	va_start(ap, max);
	n=va_arg(ap, int);
	while(n >= 0)
	{
		buf[(*pos)++]=n;
		if((*pos) >= max)	break;
		n=va_arg(ap, int);
	}
	va_end(ap);
	return(0);
}

static int cmd20_transact(int fd, unsigned char *reply, int rbytes, ...)
{
	int maxtries=2;
	unsigned char buf[16];
	int n,m=0, rpos=0, wpos=0;
	va_list ap;

	va_start(ap, rbytes);

	n=va_arg(ap, int);
	buf[0]=1;
	while(n >= 0)
	{
		buf[++m]=n;
		buf[0]++;
		n=va_arg(ap, int);
	}
	va_end(ap);

	m++;

#ifdef DEBUG
//	for(n=0; n<m; n++)
//		DEBUG_PRINT(("%d\t%02x\n", n, buf[n]));
//	DEBUG_PRINT(("\n"));
#endif

	while(wpos < m)
	{
		ssize_t w=write(fd, buf+wpos, m-wpos);
		if(w < 0)
			FAIL(CMD20_ERR_WRITE);

		wpos += w;
	}

	if(rbytes)
	while(rpos < rbytes)
	{
		ssize_t r;

		r=read(fd, reply+rpos, rbytes-rpos);
		if(r < 0)	FAIL(CMD20_ERR_READ);

		rpos += r;

		if(r == 0)
		{
			maxtries--;
			if(maxtries <= 0) break;
		}
	}

	if(rpos < rbytes)	return(CMD20_ERR_READ);

	return(0);
}

const char const *pintab[]={	"AN0", "AN1", "AN2", "AN3",
			"AN4", "AN5", "AN6", "AN7",
			"AN8", "AN9", "AN10","AN11",
			"AN12","AN13",
			"RA4",
			"P5",  "P6",  "P7",
			"RB7", "RB6", NULL };

// array index matches samples_t
const int sample_size[SAMPLES_END+1]=	{ 128,  256,  512, 1024, 2048,  4096, 8192, -1 };
const char const *sample_size_c[SAMPLES_END+1]={ "128", "256", "512", "1024", "2048", "4096", "8192", NULL };
const char const *sample_rate[RATE_END+1]=	{"1K", "2K", "4K", "8K", "20K", NULL };
const char const *cmd20_errors[CMD20_ERR_NODETECT+1]={
	"No error",
	"Read error",
	"Write error",
	"Bad parameter",
	"Sensor not ready, try later",
	"Sensor shorted",
	"Not detected",
	"Haven't tried to detect this pin" };
